"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeNames_1 = require("./TypeNames");
var UnifyClasses_1 = require("./UnifyClasses");
var REQUIRED_OVERLAP = 3 / 4;
// FIXME: Allow some type combinations to unify, like different enums,
// enums with strings, integers with doubles, maps with objects of
// the correct type.
function typeSetsCanBeCombined(s1, s2) {
    if (s1.size !== s2.size)
        return false;
    var s2ByKind = immutable_1.Map(s2.map(function (t) { return [t.kind, t]; }));
    return s1.every(function (t) {
        var kind = t.kind;
        var other = s2ByKind.get(kind);
        if (other === undefined)
            return false;
        return t.structurallyEquals(other);
    });
}
function canBeCombined(c1, c2) {
    var p1 = c1.properties;
    var p2 = c2.properties;
    if (p1.size < p2.size * REQUIRED_OVERLAP || p2.size < p1.size * REQUIRED_OVERLAP) {
        return false;
    }
    var larger;
    var smaller;
    if (p1.size > p2.size) {
        larger = p1;
        smaller = p2;
    }
    else {
        larger = p2;
        smaller = p1;
    }
    var minOverlap = Math.ceil(larger.size * REQUIRED_OVERLAP);
    var maxFaults = smaller.size - minOverlap;
    Support_1.assert(maxFaults >= 0, "Max faults negative");
    var commonProperties = [];
    var faults = 0;
    smaller.forEach(function (_, name) {
        if (larger.has(name)) {
            commonProperties.push(name);
        }
        else {
            faults += 1;
            if (faults > maxFaults)
                return false;
        }
    });
    if (faults > maxFaults)
        return false;
    for (var _i = 0, commonProperties_1 = commonProperties; _i < commonProperties_1.length; _i++) {
        var name = commonProperties_1[_i];
        var ts = smaller.get(name);
        var tl = larger.get(name);
        if (ts === undefined || tl === undefined) {
            return Support_1.panic("Both of these should have this property");
        }
        var tsCases = Type_1.nonNullTypeCases(ts);
        var tlCases = Type_1.nonNullTypeCases(tl);
        if (!tsCases.isEmpty() && !tlCases.isEmpty() && !typeSetsCanBeCombined(tsCases, tlCases)) {
            return false;
        }
    }
    return true;
}
function tryAddToClique(c, clique) {
    for (var _i = 0, _a = clique.prototypes; _i < _a.length; _i++) {
        var prototype = _a[_i];
        if (prototype.structurallyEquals(c)) {
            clique.members.push(c);
            return true;
        }
    }
    for (var _b = 0, _c = clique.prototypes; _b < _c.length; _b++) {
        var prototype = _c[_b];
        if (canBeCombined(prototype, c)) {
            clique.prototypes.push(c);
            clique.members.push(c);
            return true;
        }
    }
    return false;
}
function combineClasses(graph, stringTypeMapping) {
    var unprocessedClasses = graph
        .allNamedTypesSeparated()
        .classes.filter(function (c) { return !c.isFixed; })
        .toArray();
    var cliques = [];
    // FIXME: Don't build cliques one by one.  Instead have a list of
    // cliques-in-progress and iterate over all classes.  Add the class
    // to the first clique that it's part of.  If there's none, make it
    // into a new clique.
    while (unprocessedClasses.length > 0) {
        var classesLeft = [];
        var clique = { members: [unprocessedClasses[0]], prototypes: [unprocessedClasses[0]] };
        for (var i = 1; i < unprocessedClasses.length; i++) {
            var c = unprocessedClasses[i];
            if (!tryAddToClique(c, clique)) {
                classesLeft.push(c);
            }
        }
        if (clique.members.length > 1) {
            cliques.push(clique.members);
        }
        unprocessedClasses = classesLeft;
    }
    function makeCliqueClass(clique, builder) {
        Support_1.assert(clique.size > 0, "Clique can't be empty");
        var allNames = clique.map(function (c) { return c.getNames(); });
        return UnifyClasses_1.unifyTypes(clique, TypeNames_1.typeNamesUnion(allNames), builder, false);
    }
    return graph.rewrite(stringTypeMapping, cliques, makeCliqueClass);
}
exports.combineClasses = combineClasses;
