"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var UnifyClasses_1 = require("./UnifyClasses");
var mapSizeThreshold = 20;
function shouldBeMap(properties) {
    // Only classes with a certain number of properties are inferred
    // as maps.
    if (properties.size < mapSizeThreshold) {
        return undefined;
    }
    // We need to handle three cases for maps (and the fourth case
    // where we leave the class as is):
    //
    // 1. All property types are null.
    // 2. Some property types are null or nullable.
    // 3. No property types are null or nullable.
    var firstNonNullCases = undefined;
    var allCases = immutable_1.Set();
    var isNullable = false;
    var canBeMap = true;
    // Check that all the property types are the same, modulo nullability.
    properties.forEach(function (t) {
        // The set of types first property can be, minus null.
        var _a = Type_1.removeNullFromType(t), maybeNull = _a[0], nn = _a[1];
        if (!nn.isEmpty()) {
            if (firstNonNullCases !== undefined) {
                // The set of non-null cases for all other properties must
                // be the the same, otherwise we won't infer a map.
                if (!Type_1.unionCasesEqual(nn, firstNonNullCases, function (a, b) { return a.structurallyEquals(b); })) {
                    canBeMap = false;
                    return false;
                }
            }
            else {
                firstNonNullCases = nn;
            }
        }
        allCases = allCases.union(nn.toSet());
        if (maybeNull !== null) {
            allCases = allCases.add(maybeNull);
            isNullable = true;
        }
    });
    if (!canBeMap) {
        return undefined;
    }
    if (firstNonNullCases === undefined) {
        Support_1.assert(isNullable, "Non-nullable map candidate with no types");
    }
    return allCases;
}
function replaceClass(setOfOneClass, builder) {
    var c = Support_1.defined(setOfOneClass.first());
    var properties = c.properties;
    var shouldBe = shouldBeMap(properties);
    if (shouldBe === undefined) {
        return Support_1.panic("We shouldn't be replacing class " + c.getCombinedName() + " with a map");
    }
    // Now reconstitute all the types in the new graph.  TypeGraphs are
    // immutable, so any change in the graph actually means building a new
    // graph, and the types in the new graph are different objects.
    // Reconstituting a type means generating the "same" type in the new
    // type graph.  Except we don't get Type objects but TypeRef objects,
    // which is a type-to-be.
    return builder.getMapType(UnifyClasses_1.unifyTypes(shouldBe, c.getNames(), builder, false));
}
function inferMaps(graph, stringTypeMapping) {
    var allClasses = graph.allNamedTypesSeparated().classes;
    var classesToReplace = allClasses.filter(function (c) { return !c.isFixed && shouldBeMap(c.properties) !== undefined; }).toArray();
    return graph.rewrite(stringTypeMapping, classesToReplace.map(function (c) { return [c]; }), replaceClass);
}
exports.inferMaps = inferMaps;
