"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var TypeNames_1 = require("./TypeNames");
var UnifyClasses_1 = require("./UnifyClasses");
var PathElementKind;
(function (PathElementKind) {
    PathElementKind[PathElementKind["Root"] = 0] = "Root";
    PathElementKind[PathElementKind["Definition"] = 1] = "Definition";
    PathElementKind[PathElementKind["OneOf"] = 2] = "OneOf";
    PathElementKind[PathElementKind["AnyOf"] = 3] = "AnyOf";
    PathElementKind[PathElementKind["Property"] = 4] = "Property";
    PathElementKind[PathElementKind["AdditionalProperty"] = 5] = "AdditionalProperty";
    PathElementKind[PathElementKind["Items"] = 6] = "Items";
})(PathElementKind || (PathElementKind = {}));
function checkStringArray(arr) {
    if (!Array.isArray(arr)) {
        return Support_1.panic("Expected a string array, but got " + arr);
    }
    for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
        var e = arr_1[_i];
        if (typeof e !== "string") {
            return Support_1.panic("Expected string, but got " + e);
        }
    }
    return arr;
}
function parseRef(ref) {
    if (typeof ref !== "string") {
        return Support_1.panic("$ref must be a string");
    }
    var refName = "Something";
    var parts = ref.split("/");
    var elements = [];
    for (var i = 0; i < parts.length; i++) {
        if (parts[i] === "#") {
            elements.push({ kind: PathElementKind.Root });
            refName = "Root";
        }
        else if (parts[i] === "definitions" && i + 1 < parts.length) {
            refName = parts[i + 1];
            elements.push({ kind: PathElementKind.Definition, name: refName });
            i += 1;
        }
        else {
            Support_1.panic("Could not parse JSON schema reference " + ref);
        }
    }
    return [immutable_1.List(elements), refName];
}
function lookupDefinition(schema, name) {
    var definitions = Support_1.checkStringMap(schema.definitions);
    return Support_1.checkStringMap(definitions[name]);
}
function lookupProperty(schema, name) {
    var properties = Support_1.checkStringMap(schema.properties);
    return Support_1.checkStringMap(properties[name]);
}
function indexArray(cases, index) {
    if (!Array.isArray(cases)) {
        return Support_1.panic("oneOf or anyOf value must be an array");
    }
    return Support_1.checkStringMap(cases[index]);
}
function getName(schema, typeNames) {
    if (!typeNames.areInferred) {
        return typeNames;
    }
    var title = schema.title;
    if (typeof title === "string") {
        return TypeNames_1.makeTypeNames(title, false);
    }
    else {
        return typeNames.makeInferred();
    }
}
function checkTypeList(typeOrTypes) {
    if (typeof typeOrTypes === "string") {
        return immutable_1.OrderedSet([typeOrTypes]);
    }
    else if (Array.isArray(typeOrTypes)) {
        var arr = [];
        for (var _i = 0, typeOrTypes_1 = typeOrTypes; _i < typeOrTypes_1.length; _i++) {
            var t = typeOrTypes_1[_i];
            if (typeof t !== "string") {
                return Support_1.panic("element of type is not a string: " + t);
            }
            arr.push(t);
        }
        var set = immutable_1.OrderedSet(arr);
        Support_1.assert(!set.isEmpty(), "JSON Schema must specify at least one type");
        return set;
    }
    else {
        return Support_1.panic("type is neither a string or array of strings: " + typeOrTypes);
    }
}
function makeImmutablePath(path) {
    return path.map(function (pe) { return immutable_1.fromJS(pe); });
}
function schemaToType(typeBuilder, topLevelName, rootJson) {
    var root = Support_1.checkStringMap(rootJson);
    var typeForPath = immutable_1.Map();
    function setTypeForPath(path, t) {
        typeForPath = typeForPath.set(makeImmutablePath(path), t);
    }
    function lookupRef(local, localPath, ref) {
        var first = ref.first();
        if (first === undefined) {
            return [local, localPath];
        }
        var rest = ref.rest();
        if (first.kind === PathElementKind.Root) {
            return lookupRef(root, immutable_1.List([first]), ref.rest());
        }
        localPath = localPath.push(first);
        switch (first.kind) {
            case PathElementKind.Definition:
                return lookupRef(lookupDefinition(local, first.name), localPath, rest);
            case PathElementKind.OneOf:
                return lookupRef(indexArray(local.oneOf, first.index), localPath, rest);
            case PathElementKind.AnyOf:
                return lookupRef(indexArray(local.anyOf, first.index), localPath, rest);
            case PathElementKind.Property:
                return lookupRef(lookupProperty(local, first.name), localPath, rest);
            case PathElementKind.AdditionalProperty:
                return lookupRef(Support_1.checkStringMap(local.additionalProperties), localPath, rest);
            case PathElementKind.Items:
                return lookupRef(Support_1.checkStringMap(local.items), localPath, rest);
            default:
                return Support_1.assertNever(first);
        }
    }
    function makeClass(schema, path, typeNames, properties, requiredArray) {
        var required = immutable_1.Set(requiredArray);
        var result = typeBuilder.getUniqueClassType(getName(schema, typeNames));
        var c = Type_1.assertIsClass(TypeBuilder_1.getHopefullyFinishedType(typeBuilder, result));
        setTypeForPath(path, result);
        // FIXME: We're using a Map instead of an OrderedMap here because we represent
        // the JSON Schema as a JavaScript object, which has no map ordering.  Ideally
        // we would use a JSON parser that preserves order.
        var props = immutable_1.Map(properties).map(function (propSchema, propName) {
            var t = toType(Support_1.checkStringMap(propSchema), path.push({ kind: PathElementKind.Property, name: propName }), TypeNames_1.makeTypeNames(pluralize.singular(propName), true));
            if (!required.has(propName)) {
                return typeBuilder.makeNullable(t, TypeNames_1.makeTypeNames(propName, true));
            }
            return t;
        });
        c.setProperties(props.toOrderedMap());
        return result;
    }
    function makeMap(path, typeNames, additional) {
        var valuesType = undefined;
        var mustSet = false;
        var result = typeBuilder.getLazyMapType(function () {
            mustSet = true;
            return valuesType;
        });
        setTypeForPath(path, result);
        path = path.push({ kind: PathElementKind.AdditionalProperty });
        valuesType = toType(additional, path, typeNames.singularize());
        if (mustSet) {
            result.deref()[0].setValues(valuesType);
        }
        return result;
    }
    function fromTypeName(schema, path, typeNames, typeName) {
        typeNames = getName(schema, typeNames.makeInferred());
        switch (typeName) {
            case "object":
                var required = void 0;
                if (schema.required === undefined) {
                    required = [];
                }
                else {
                    required = checkStringArray(schema.required);
                }
                if (schema.properties !== undefined) {
                    return makeClass(schema, path, typeNames, Support_1.checkStringMap(schema.properties), required);
                }
                else if (schema.additionalProperties !== undefined) {
                    var additional = schema.additionalProperties;
                    if (additional === true) {
                        return typeBuilder.getMapType(typeBuilder.getPrimitiveType("any"));
                    }
                    else if (additional === false) {
                        return makeClass(schema, path, typeNames, {}, required);
                    }
                    else {
                        return makeMap(path, typeNames, Support_1.checkStringMap(additional));
                    }
                }
                else {
                    return typeBuilder.getMapType(typeBuilder.getPrimitiveType("any"));
                }
            case "array":
                if (schema.items !== undefined) {
                    path = path.push({ kind: PathElementKind.Items });
                    return typeBuilder.getArrayType(toType(Support_1.checkStringMap(schema.items), path, typeNames.singularize()));
                }
                return typeBuilder.getArrayType(typeBuilder.getPrimitiveType("any"));
            case "boolean":
                return typeBuilder.getPrimitiveType("bool");
            case "string":
                if (schema.format !== undefined) {
                    switch (schema.format) {
                        case "date":
                            return typeBuilder.getPrimitiveType("date");
                        case "time":
                            return typeBuilder.getPrimitiveType("time");
                        case "date-time":
                            return typeBuilder.getPrimitiveType("date-time");
                        default:
                            return Support_1.panic("String format " + schema.format + " not supported");
                    }
                }
                return typeBuilder.getStringType(typeNames, undefined);
            case "null":
                return typeBuilder.getPrimitiveType("null");
            case "integer":
                return typeBuilder.getPrimitiveType("integer");
            case "number":
                return typeBuilder.getPrimitiveType("double");
            default:
                return Support_1.panic("not a type name: " + typeName);
        }
    }
    function convertToType(schema, path, typeNames) {
        typeNames = getName(schema, typeNames);
        function convertOneOrAnyOf(cases, kind) {
            if (!Array.isArray(cases)) {
                return Support_1.panic("oneOf or anyOf is not an array: " + cases);
            }
            // FIXME: This cast shouldn't be necessary, but TypeScript forces our hand.
            var types = cases.map(function (t, index) { return toType(Support_1.checkStringMap(t), path.push({ kind: kind, index: index }), typeNames).deref()[0]; });
            return UnifyClasses_1.unifyTypes(immutable_1.OrderedSet(types), typeNames, typeBuilder, true);
        }
        if (schema.$ref !== undefined) {
            var _a = parseRef(schema.$ref), ref = _a[0], refName = _a[1];
            var _b = lookupRef(schema, path, ref), target = _b[0], targetPath = _b[1];
            return toType(target, targetPath, typeNames.areInferred ? TypeNames_1.makeTypeNames(refName, true) : typeNames);
        }
        else if (schema.enum !== undefined) {
            return typeBuilder.getEnumType(typeNames, immutable_1.OrderedSet(checkStringArray(schema.enum)));
        }
        else if (schema.type !== undefined) {
            var jsonTypes = checkTypeList(schema.type);
            if (jsonTypes.size === 1) {
                return fromTypeName(schema, path, typeNames, Support_1.defined(jsonTypes.first()));
            }
            else {
                var types = jsonTypes.map(function (n) { return fromTypeName(schema, path, typeNames, n).deref()[0]; });
                return UnifyClasses_1.unifyTypes(types, typeNames, typeBuilder, true);
            }
        }
        else if (schema.oneOf !== undefined) {
            return convertOneOrAnyOf(schema.oneOf, PathElementKind.OneOf);
        }
        else if (schema.anyOf !== undefined) {
            return convertOneOrAnyOf(schema.anyOf, PathElementKind.AnyOf);
        }
        else {
            return typeBuilder.getPrimitiveType("any");
        }
    }
    function toType(schema, path, typeNames) {
        // FIXME: This fromJS thing is ugly and inefficient.  Schemas aren't
        // big, so it most likely doesn't matter.
        var immutablePath = makeImmutablePath(path);
        var maybeType = typeForPath.get(immutablePath);
        if (maybeType !== undefined) {
            return maybeType;
        }
        var result = convertToType(schema, path, typeNames);
        setTypeForPath(immutablePath, result);
        return result;
    }
    var rootPathElement = { kind: PathElementKind.Root };
    var rootType = toType(root, immutable_1.List([rootPathElement]), TypeNames_1.makeTypeNames(topLevelName, false));
    return rootType;
}
exports.schemaToType = schemaToType;
