"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Naming_1 = require("./Naming");
var Source_1 = require("./Source");
var Annotation_1 = require("./Annotation");
var Support_1 = require("./Support");
function lineIndentation(line) {
    var len = line.length;
    var indent = 0;
    for (var i = 0; i < len; i++) {
        var c = line.charAt(i);
        if (c === " ") {
            indent += 1;
        }
        else if (c === "\t") {
            indent = (indent / 4 + 1) * 4;
        }
        else {
            return { indent: indent, text: line.substring(i) };
        }
    }
    return { indent: 0, text: null };
}
var Renderer = /** @class */ (function () {
    function Renderer(typeGraph, leadingComments) {
        var _this = this;
        this.typeGraph = typeGraph;
        this.leadingComments = leadingComments;
        this.emitNewline = function () {
            var nl = Source_1.newline();
            _this._currentEmitTarget.push(nl);
            _this._lastNewline = nl;
        };
        this.emitItem = function (item) {
            if (_this._needBlankLine) {
                _this.emitNewline();
                _this._needBlankLine = false;
            }
            _this._currentEmitTarget.push(item);
        };
        this.ensureBlankLine = function () {
            if (_this._emitted.length === 0 && _this._currentEmitTarget.length === 0) {
                // no blank lines at start of file
                return;
            }
            _this._needBlankLine = true;
        };
        this.emitTable = function (tableArray) {
            var table = immutable_1.List(tableArray.map(function (r) { return immutable_1.List(r.map(function (sl) { return Source_1.sourcelikeToSource(sl); })); }));
            _this.emitItem({ kind: "table", table: table });
            _this.emitNewline();
        };
        this.finishedSource = function () {
            return Source_1.sourcelikeToSource(_this._emitted);
        };
        this.render = function () {
            _this._names = Naming_1.assignNames(immutable_1.OrderedSet(_this.setUpNaming()));
            _this.emitSource();
            return { rootSource: _this.finishedSource(), names: _this._names };
        };
        this._currentEmitTarget = this._emitted = [];
        this._needBlankLine = false;
    }
    Renderer.prototype.emitLine = function () {
        var lineParts = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            lineParts[_i] = arguments[_i];
        }
        if (lineParts.length === 1) {
            this.emitItem(lineParts[0]);
        }
        else if (lineParts.length > 1) {
            this.emitItem(lineParts);
        }
        this.emitNewline();
    };
    Renderer.prototype.emitMultiline = function (linesString) {
        var lines = linesString.split("\n");
        var numLines = lines.length;
        if (numLines === 0)
            return;
        this.emitLine(lines[0]);
        var currentIndent = 0;
        for (var i = 1; i < numLines; i++) {
            var line = lines[i];
            var _a = lineIndentation(line), indent = _a.indent, text = _a.text;
            Support_1.assert(indent % 4 === 0, "Indentation is not a multiple of 4.");
            if (text !== null) {
                var newIndent = indent / 4;
                this.changeIndent(newIndent - currentIndent);
                currentIndent = newIndent;
                this.emitLine(text);
            }
            else {
                this.emitNewline();
            }
        }
        if (currentIndent !== 0) {
            this.changeIndent(-currentIndent);
        }
    };
    Renderer.prototype.emitAnnotated = function (annotation, emitter) {
        var oldEmitTarget = this._currentEmitTarget;
        var emitTarget = [];
        this._currentEmitTarget = emitTarget;
        emitter();
        Support_1.assert(this._currentEmitTarget === emitTarget, "_currentEmitTarget not restored correctly");
        this._currentEmitTarget = oldEmitTarget;
        var source = Source_1.sourcelikeToSource(emitTarget);
        this._currentEmitTarget.push(Source_1.annotated(annotation, source));
    };
    Renderer.prototype.emitIssue = function (message, emitter) {
        this.emitAnnotated(new Annotation_1.IssueAnnotationData(message), emitter);
    };
    Renderer.prototype.changeIndent = function (offset) {
        if (!this._lastNewline) {
            return Support_1.panic("Cannot change indent for the first line");
        }
        this._lastNewline.indentationChange += offset;
    };
    Renderer.prototype.forEach = function (iterable, interposedBlankLines, leadingBlankLine, emitter) {
        var _this = this;
        var onFirst = true;
        iterable.forEach(function (v, k) {
            if ((leadingBlankLine && onFirst) || (interposedBlankLines && !onFirst)) {
                _this.ensureBlankLine();
            }
            emitter(v, k);
            onFirst = false;
        });
    };
    Renderer.prototype.forEachWithBlankLines = function (iterable, blankLineLocations, emitter) {
        var interposing = ["interposing", "leading-and-interposing"].indexOf(blankLineLocations) >= 0;
        var leading = ["leading", "leading-and-interposing"].indexOf(blankLineLocations) >= 0;
        this.forEach(iterable, interposing, leading, emitter);
    };
    Renderer.prototype.indent = function (fn) {
        this.changeIndent(1);
        fn();
        this.changeIndent(-1);
    };
    Object.defineProperty(Renderer.prototype, "names", {
        get: function () {
            if (this._names === undefined) {
                return Support_1.panic("Names accessed before they were assigned");
            }
            return this._names;
        },
        enumerable: true,
        configurable: true
    });
    return Renderer;
}());
exports.Renderer = Renderer;
