import { OrderedSet, OrderedMap, Collection } from "immutable";
import { TypeRef, TypeReconstituter } from "./TypeBuilder";
import { TypeNames } from "./TypeNames";
export declare type PrimitiveStringTypeKind = "string" | "date" | "time" | "date-time";
export declare type PrimitiveTypeKind = "any" | "null" | "bool" | "integer" | "double" | PrimitiveStringTypeKind;
export declare type NamedTypeKind = "class" | "enum" | "union";
export declare type TypeKind = PrimitiveTypeKind | NamedTypeKind | "array" | "map";
export declare abstract class Type {
    readonly typeRef: TypeRef;
    readonly kind: TypeKind;
    constructor(typeRef: TypeRef, kind: TypeKind);
    readonly abstract children: OrderedSet<Type>;
    directlyReachableTypes<T>(setForType: (t: Type) => OrderedSet<T> | null): OrderedSet<T>;
    readonly hasNames: boolean;
    getNames(): TypeNames;
    getCombinedName(): string;
    getProposedNames(): OrderedSet<string>;
    readonly abstract isNullable: boolean;
    abstract isPrimitive(): this is PrimitiveType;
    abstract map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    equals(other: any): boolean;
    hashCode(): number;
    protected abstract structuralEqualityStep(other: Type, queue: (a: Type, b: Type) => boolean): boolean;
    structurallyEquals(other: Type): boolean;
}
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    constructor(typeRef: TypeRef, kind: PrimitiveTypeKind, checkKind?: boolean);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, _: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(_other: Type, _queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class StringType extends PrimitiveType {
    readonly enumCases: OrderedMap<string, number> | undefined;
    constructor(typeRef: TypeRef, enumCases: OrderedMap<string, number> | undefined);
    map(builder: TypeReconstituter, _: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(_other: Type, _queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class ArrayType extends Type {
    private _itemsRef;
    readonly kind: "array";
    constructor(typeRef: TypeRef, _itemsRef?: TypeRef | undefined);
    setItems(itemsRef: TypeRef): undefined;
    private getItemsRef();
    readonly items: Type;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: ArrayType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class MapType extends Type {
    private _valuesRef;
    readonly kind: "map";
    constructor(typeRef: TypeRef, _valuesRef?: TypeRef | undefined);
    setValues(valuesRef: TypeRef): undefined;
    private getValuesRef();
    readonly values: Type;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: MapType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class ClassType extends Type {
    readonly isFixed: boolean;
    private _propertyRefs;
    kind: "class";
    private _propertyTypes;
    constructor(typeRef: TypeRef, isFixed: boolean, _propertyRefs?: OrderedMap<string, TypeRef> | undefined);
    setProperties(propertyRefs: OrderedMap<string, TypeRef>): void;
    private getPropertyRefs();
    readonly properties: OrderedMap<string, Type>;
    readonly sortedProperties: OrderedMap<string, Type>;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: ClassType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare function assertIsClass(t: Type): ClassType;
export declare class EnumType extends Type {
    readonly cases: OrderedSet<string>;
    kind: "enum";
    constructor(typeRef: TypeRef, cases: OrderedSet<string>);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, _: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: EnumType, _queue: (a: Type, b: Type) => void): boolean;
}
export declare class UnionType extends Type {
    private _memberRefs;
    kind: "union";
    constructor(typeRef: TypeRef, _memberRefs?: OrderedSet<TypeRef> | undefined);
    setMembers(memberRefs: OrderedSet<TypeRef>): undefined;
    private getMemberRefs();
    readonly members: OrderedSet<Type>;
    readonly stringTypeMembers: OrderedSet<Type>;
    findMember(kind: TypeKind): Type | undefined;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    readonly sortedMembers: OrderedSet<Type>;
    protected structuralEqualityStep(other: UnionType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare function unionCasesEqual(ma: OrderedSet<Type>, mb: OrderedSet<Type>, membersEqual: (a: Type, b: Type) => boolean): boolean;
export declare function removeNullFromUnion(t: UnionType): [PrimitiveType | null, OrderedSet<Type>];
export declare function removeNullFromType(t: Type): [PrimitiveType | null, OrderedSet<Type>];
export declare function nullableFromUnion(t: UnionType): Type | null;
export declare function nonNullTypeCases(t: Type): OrderedSet<Type>;
export declare function isNamedType(t: Type): boolean;
export declare type SeparatedNamedTypes = {
    classes: OrderedSet<ClassType>;
    enums: OrderedSet<EnumType>;
    unions: OrderedSet<UnionType>;
};
export declare function separateNamedTypes(types: Collection<any, Type>): SeparatedNamedTypes;
export declare type StringTypeMatchers<U> = {
    dateType?: (dateType: PrimitiveType) => U;
    timeType?: (timeType: PrimitiveType) => U;
    dateTimeType?: (dateTimeType: PrimitiveType) => U;
};
export declare function matchTypeExhaustive<U>(t: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: StringType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U, dateType: (dateType: PrimitiveType) => U, timeType: (timeType: PrimitiveType) => U, dateTimeType: (dateTimeType: PrimitiveType) => U): U;
export declare function matchType<U>(t: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: StringType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U, stringTypeMatchers?: StringTypeMatchers<U>): U;
export declare function matchCompoundType(t: Type, arrayType: (arrayType: ArrayType) => void, classType: (classType: ClassType) => void, mapType: (mapType: MapType) => void, unionType: (unionType: UnionType) => void): void;
