import { Map, OrderedMap, OrderedSet, Set } from "immutable";
import { PrimitiveTypeKind, Type, PrimitiveStringTypeKind } from "./Type";
import { TypeGraph } from "./TypeGraph";
import { TypeNames } from "./TypeNames";
export declare type TypeRefCallback = (index: number) => void;
export declare class TypeRef {
    readonly graph: TypeGraph;
    private _allocatingTypeBuilder;
    private _maybeIndexOrRef?;
    private _callbacks?;
    constructor(graph: TypeGraph, index?: number, _allocatingTypeBuilder?: TypeBuilder | undefined);
    private follow();
    readonly maybeIndex: number | undefined;
    readonly index: number;
    callWhenResolved(callback: TypeRefCallback): void;
    resolve(tref: TypeRef): void;
    deref(): [Type, TypeNames | undefined];
    equals(other: any): boolean;
    hashCode(): number;
}
export declare type StringTypeMapping = {
    date: PrimitiveStringTypeKind;
    time: PrimitiveStringTypeKind;
    dateTime: PrimitiveStringTypeKind;
};
export declare abstract class TypeBuilder {
    private readonly _stringTypeMapping;
    readonly typeGraph: TypeGraph;
    protected topLevels: Map<string, TypeRef>;
    protected readonly types: (Type | undefined)[];
    private readonly typeNames;
    constructor(_stringTypeMapping: StringTypeMapping);
    addTopLevel(name: string, tref: TypeRef): void;
    reserveTypeRef(): TypeRef;
    private commitType;
    protected addType<T extends Type>(forwardingRef: TypeRef | undefined, creator: (tref: TypeRef) => T, names: TypeNames | undefined): TypeRef;
    atIndex: (index: number) => [Type, TypeNames | undefined];
    addNames: (tref: TypeRef, names: TypeNames) => void;
    makeNullable: (tref: TypeRef, typeNames: TypeNames) => TypeRef;
    finish(): TypeGraph;
    private _primitiveTypes;
    private _noEnumStringType;
    private _mapTypes;
    private _arrayTypes;
    private _enumTypes;
    private _classTypes;
    private _unionTypes;
    getPrimitiveType(kind: PrimitiveTypeKind, forwardingRef?: TypeRef): TypeRef;
    getStringType(names: TypeNames | undefined, cases: OrderedMap<string, number> | undefined, forwardingRef?: TypeRef): TypeRef;
    getEnumType(names: TypeNames, cases: OrderedSet<string>, forwardingRef?: TypeRef): TypeRef;
    getMapType(values: TypeRef, forwardingRef?: TypeRef): TypeRef;
    getArrayType(items: TypeRef, forwardingRef?: TypeRef): TypeRef;
    getClassType(names: TypeNames, properties: OrderedMap<string, TypeRef>, forwardingRef?: TypeRef): TypeRef;
    getUniqueClassType: (names: TypeNames, properties?: OrderedMap<string, TypeRef> | undefined, forwardingRef?: TypeRef | undefined) => TypeRef;
    getUnionType(names: TypeNames, members: OrderedSet<TypeRef>, forwardingRef?: TypeRef): TypeRef;
    getUniqueUnionType: (names: TypeNames, members: OrderedSet<TypeRef>) => TypeRef;
}
export interface TypeLookerUp {
    lookupTypeRefs(typeRefs: TypeRef[]): TypeRef | undefined;
    lookupTypeRef(typeRef: TypeRef): TypeRef;
    lookupType(typeRef: TypeRef): Type | undefined;
}
export declare function getHopefullyFinishedType(builder: TypeLookerUp, t: TypeRef): Type;
export declare class TypeGraphBuilder extends TypeBuilder implements TypeLookerUp {
    protected typeForEntry(entry: Type | undefined): Type | undefined;
    getLazyMapType(valuesCreator: () => TypeRef | undefined): TypeRef;
    lookupTypeRefs(_typeRefs: TypeRef[]): undefined;
    lookupTypeRef: (typeRef: TypeRef) => TypeRef;
    lookupType: (typeRef: TypeRef) => Type | undefined;
}
export declare class TypeReconstituter {
    private readonly _typeBuilder;
    private readonly _typeNames;
    private readonly _forwardingRef;
    private _wasUsed;
    constructor(_typeBuilder: TypeBuilder, _typeNames: TypeNames | undefined, _forwardingRef: TypeRef);
    private useBuilder;
    getPrimitiveType: (kind: PrimitiveTypeKind) => TypeRef;
    getStringType: (enumCases: OrderedMap<string, number> | undefined) => TypeRef;
    getEnumType: (cases: OrderedSet<string>) => TypeRef;
    getMapType: (values: TypeRef) => TypeRef;
    getArrayType: (items: TypeRef) => TypeRef;
    getClassType: (properties: OrderedMap<string, TypeRef>) => TypeRef;
    getUniqueClassType: (properties?: OrderedMap<string, TypeRef> | undefined) => TypeRef;
    getUnionType: (members: OrderedSet<TypeRef>) => TypeRef;
}
export declare class GraphRewriteBuilder<T extends Type> extends TypeBuilder implements TypeLookerUp {
    private readonly _originalGraph;
    private readonly _replacer;
    private _setsToReplaceByMember;
    private _reconstitutedTypes;
    constructor(_originalGraph: TypeGraph, stringTypeMapping: StringTypeMapping, setsToReplace: T[][], _replacer: (typesToReplace: Set<T>, builder: GraphRewriteBuilder<T>) => TypeRef);
    followIndex(index: number): number;
    protected typeForEntry(entry: Type | undefined | number): Type | undefined;
    private withForwardingRef(typeCreator);
    private replaceSet(typesToReplace);
    private getReconstitutedType;
    reconstituteType: (t: Type) => TypeRef;
    lookupTypeRefs(typeRefs: TypeRef[]): TypeRef | undefined;
    lookupTypeRef: (typeRef: TypeRef) => TypeRef;
    lookupType: (typeRef: TypeRef) => Type | undefined;
    finish(): TypeGraph;
}
export declare abstract class UnionBuilder<TBuilder extends TypeBuilder, TArray, TClass, TMap> {
    protected readonly typeBuilder: TBuilder;
    protected readonly typeNames: TypeNames;
    private _haveAny;
    private _haveNull;
    private _haveBool;
    private _haveInteger;
    private _haveDouble;
    private _stringTypes;
    private readonly _arrays;
    private readonly _maps;
    private readonly _classes;
    private _enumCaseMap;
    private _enumCases;
    constructor(typeBuilder: TBuilder, typeNames: TypeNames);
    readonly haveString: boolean;
    addAny: () => void;
    addNull: () => void;
    addBool: () => void;
    addInteger: () => void;
    addDouble: () => void;
    addStringType: (kind: PrimitiveStringTypeKind) => void;
    addArray: (t: TArray) => void;
    addClass: (t: TClass) => void;
    addMap: (t: TMap) => void;
    addEnumCase: (s: string, count?: number) => void;
    addEnumCases: (cases: OrderedMap<string, number>) => void;
    protected abstract makeEnum(cases: string[], counts: {
        [name: string]: number;
    }): TypeRef;
    protected abstract makeClass(classes: TClass[], maps: TMap[]): TypeRef;
    protected abstract makeArray(arrays: TArray[]): TypeRef;
    buildUnion: (unique: boolean) => TypeRef;
}
