"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var TypeGraph = /** @class */ (function () {
    function TypeGraph(typeBuilder) {
        var _this = this;
        // FIXME: OrderedMap?  We lose the order in PureScript right now, though,
        // and maybe even earlier in the TypeScript driver.
        this._topLevels = immutable_1.Map();
        this.freeze = function (topLevels, types, typeNames) {
            Support_1.assert(!_this.isFrozen, "Tried to freeze TypeGraph a second time");
            Support_1.assert(types.every(function (t) { return t.typeRef.graph === _this; }), "Trying to freeze a graph with types that don't belong in it");
            // The order of these three statements matters.  If we set _typeBuilder
            // to undefined before we deref the TypeRefs, then we need to set _types
            // before, also, because the deref will call into typeAtIndex, which requires
            // either a _typeBuilder or a _types.
            _this._types = types;
            _this._typeNames = typeNames;
            _this._typeBuilder = undefined;
            _this._topLevels = topLevels.map(function (tref) { return tref.deref()[0]; });
        };
        this.allNamedTypes = function (childrenOfType) {
            return _this.filterTypes(Type_1.isNamedType, childrenOfType);
        };
        this.allNamedTypesSeparated = function (childrenOfType) {
            var types = _this.allNamedTypes(childrenOfType);
            return Type_1.separateNamedTypes(types);
        };
        this.allTypesUnordered = function () {
            Support_1.assert(_this.isFrozen, "Tried to get all graph types before it was frozen");
            return immutable_1.Set(Support_1.defined(_this._types));
        };
        this._typeBuilder = typeBuilder;
    }
    Object.defineProperty(TypeGraph.prototype, "isFrozen", {
        get: function () {
            return this._typeBuilder === undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeGraph.prototype, "topLevels", {
        get: function () {
            Support_1.assert(this.isFrozen, "Cannot get top-levels from a non-frozen graph");
            return Support_1.defined(this._topLevels);
        },
        enumerable: true,
        configurable: true
    });
    TypeGraph.prototype.atIndex = function (index) {
        if (this._typeBuilder !== undefined) {
            return this._typeBuilder.atIndex(index);
        }
        return [Support_1.defined(this._types)[index], Support_1.defined(this._typeNames)[index]];
    };
    TypeGraph.prototype.filterTypes = function (predicate, childrenOfType) {
        var seen = immutable_1.Set();
        var types = immutable_1.List();
        function addFromType(t) {
            if (seen.has(t))
                return;
            seen = seen.add(t);
            var children = childrenOfType ? childrenOfType(t) : t.children;
            children.forEach(addFromType);
            if (predicate(t)) {
                types = types.push(t);
            }
        }
        this.topLevels.forEach(addFromType);
        return types.reverse().toOrderedSet();
    };
    // Each array in `replacementGroups` is a bunch of types to be replaced by a
    // single new type.  `replacer` is a function that takes a group and a
    // TypeBuilder, and builds a new type with that builder that replaces the group.
    // That particular TypeBuilder will have to take as inputs types in the old
    // graph, but return types in the new graph.  Recursive types must be handled
    // carefully.
    TypeGraph.prototype.rewrite = function (stringTypeMapping, replacementGroups, replacer) {
        if (replacementGroups.length === 0)
            return this;
        return new TypeBuilder_1.GraphRewriteBuilder(this, stringTypeMapping, replacementGroups, replacer).finish();
    };
    return TypeGraph;
}());
exports.TypeGraph = TypeGraph;
